#define FILE_NAME "AvxOmp.c"

#if defined(PDEF_H)

 LIST("# " FILE_NAME EJ(" version 8.16"," バージョン 8.16"))
 INTF(N,4,SZ,   "size of matrix")
 #if defined(_OPENMP)
  INTF(Z,5,500,  "iteration number")
 #else
  INTF(Z,5,100,  "iteration number")
 #endif
 FMAT(S,307,N,N,"Original matrix")
 FMAT(A,307,N,N,"Main matrix(A=A*C)")
 FMAT(C,307,N,N,"Copy of A  (A=A*C)")
 FMAT(T,307,N,N,"Transposed matrix of C")
 PROC(Gen,"Generate start matrix")
 AUTO("Gen")
 #if defined(_OPENMP)
  PROC(F87Mth1,"f87 multi thread(1)")
  AUTO("F87Mth1")
  PROC(F87Mth2,"f87 multi thread(2)")
  AUTO("F87Mth2")
  PROC(F87Mth3,"f87 multi thread(3)")
  AUTO("F87Mth3")
 #else
  PROC(F87Sth1,"f87 single thread(1)")
  AUTO("F87Sth1")
  PROC(F87Sth2,"f87 single thread(2)")
  AUTO("F87Sth2")
 #endif
 #if SSE
  #if defined(_OPENMP)
   PROC(SseMth1,"SSE multi thread(1)")
   AUTO("SseMth1")
   PROC(SseMth2,"SSE multi thread(2)")
   AUTO("SseMth2")
  #else
   PROC(SseSth,"SSE single thread")
   AUTO("SseSth")
  #endif
 #endif
 #if AVX
  #if defined(_OPENMP)
   PROC(AvxMth1,"AVX multi thread(1)")
   AUTO("AvxMth1")
   PROC(AvxMth2,"AVX multi thread(2)")
   AUTO("AvxMth2")
  #else
   PROC(AvxSth,"AVX single thread")
   AUTO("AvxSth")
  #endif
 #endif

#else

 #if defined(__GNUC__)
  #define SSE  1
  #define AVX  1
 #elif defined(_MSC_VER)
  #define SSE  (_MSC_VER>=1300)
  #define AVX  (_MSC_VER>=1600)
 #else
  #define SSE  0
  #define AVX  0
 #endif
 #if AVX
  #define ALIGN_MEM 32
 #elif SSE
  #define ALIGN_MEM 16
 #else
  #define ALIGN_MEM  8
 #endif
 #define X86_INTRIN

 /* 行列の大きさ */
 #define SZ   512

 /* ポコムバージョン 8.01 以降で静的配列モードのために定義する */
 #define MROW SZ
 #define MCOL SZ

 #include "pocom.h"
 CHECK_FILE_NAME

 #if AVX
  #include <immintrin.h>
 #elif SSE
  #include <xmmintrin.h> 
 #endif

 void Gen(void) {
     #define COEF 0.333666
     Int i; 
     NoList=1;
     Print("N Z");
     Alloc("T A C S");
 #if defined(_OPENMP) && 1
     {
         /* 最も高速だがスレッド数によって結果が安定しない */
         int num_procs=omp_get_num_procs();
         Rand_t*r=(Rand_t*)GetMem(num_procs,sizeof*r);
     #pragma omp parallel for
         for (i=0;i<num_procs;i++) Srand_r(i,r[i]);
     #pragma omp parallel for
         for (i=0;i<N;i++) {
             Int j; int thread_num=omp_get_thread_num();
             for (j=0;j<N;j++)
                 S[i][j]=(float)(Unif_r(r[thread_num])*COEF-0.5*COEF);
         }
         FreeMem(r);
     }
 #elif defined(_OPENMP) && 1
         /* 次に高速で結果が安定している */
     #pragma omp parallel for
     for (i=0;i<N;i++) {
         Int j; Rand_t r; Srand_r(i,r);
         for (j=0;j<N;j++) S[i][j]=(float)(Unif_r(r)*COEF-0.5*COEF);
     }
 #elif defined(_OPENMP) && 1
     /* １個の乱数発生器を複数スレッドで共有していので、
        クリティカルセクションの負荷がかかり非常に遅くなる。
        しかも結果が安定していない */
     #pragma omp parallel for
     for (i=0;i<N;i++) {
         Int j;
         for (j=0;j<N;j++) S[i][j]=(float)(Unif()*COEF-0.5*COEF);
     }
 #else
     /* OpenMP を使用しない。上の方法より速い。結果が安定している */
     for (i=0;i<N;i++) {
         Int j;
         for (j=0;j<N;j++) S[i][j]=(float)(Unif()*COEF-0.5*COEF);
     }
 #endif
     Lap("Gen()");
 }

 void Copy(void) {
    Int i,j;
    for (i=0;i<N;i++) for (j=0;j<N;j++) C[i][j]=T[j][i]=A[i][j]=S[i][j];
 }

 void PrintSum(const char*msg) {
    Int i,j; float sum=0; char buf[200];
    for (i=0;i<N;i++) for (j=0;j<N;j++) sum+=A[i][j];
    sprintf(buf,"%s sum=%8.3f",msg,sum);
    Lap(buf);
 }

 void F87(Int i,Int j) {
    Int k; fvec left=C[i],right=T[j]; float sum=0;
    for (k=0;k<N;k++) sum+=(float)(left[k]*right[k]);
    A[i][j]=sum;
 }

 #if defined(_OPENMP)
 
 void F87Mth1(void) {
    Int i,z;
    Copy();
    for (z=0;z<Z;z++) {
        #pragma omp parallel for
        for (i=0;i<N;i++) {
            Int j;
            for (j=0;j<N;j++) F87(i,j);
        }
        Progress("#F87Mth(1) ",z,Z);
    }
    PrintSum("Normal multi  thread(1)");
 }

 void F87Mth2(void) {
    Int z,n=0;
    Copy();
    #pragma omp parallel for
    for (z=0;z<Z;z++) {
        Int i,j;
        for (i=0;i<N;i++) for (j=0;j<N;j++) F87(i,j);
        #pragma omp critical
        Progress("#F87Mth(2) ",n++,Z);
    }
    PrintSum("Normal multi  thread(2)");
 }

 void F87Mth3(void) {
    Int z,n=0,m=0;
    Copy();
    #pragma omp parallel for
    for (z=0;z<Z;z++) {
        Int i,j;
        for (i=0;i<N;i++) {
            for (j=0;j<N;j++) F87(i,j);
            #pragma omp critical
            Progress2("#F87Mth(3) ",n,Z,m++,N);
        }
        #pragma omp atomic
        n++;
        m=0;
    }
    PrintSum("Normal multi  thread(3)");
 }
 #else
 void F87Sth1(void) {
    Int i,j,z;
    Copy();
    for (z=0;z<Z;z++) {
        for (i=0;i<N;i++) for (j=0;j<N;j++) F87(i,j);
        Progress("#F87Sth1 ",z,Z);
    }
    PrintSum("Normal single thread(1)");
 }
 void F87Sth2(void) {
    Int i,j,z;
    Copy();
    for (z=0;z<Z;z++) {
        for (i=0;i<N;i++) {
            for (j=0;j<N;j++) F87(i,j);
            Progress2("#F87Sth2 ",z,Z,i,N);
        }
    }
    PrintSum("Normal single thread(2)");
 }
 #endif

 #if SSE
  #if defined(__GNUC__)
   #pragma GCC push_options
   #pragma GCC target("sse")
  #endif
 void Sse(Int i,Int j) {
     Int k; fvec left=C[i],right=T[j]; float u[4]; __m128 m;
     m=_mm_setzero_ps();
     for (k=0;k<N;k+=4) 
         m=_mm_add_ps(m,_mm_mul_ps(*(__m128*)&left[k],*(__m128*)&right[k]));
     _mm_storeu_ps(u,m);
     A[i][j]=(float)(u[0]+u[1]+u[2]+u[3]);
 }
  #if defined(__GNUC__)
   #pragma GCC pop_options
  #endif

  #if defined(_OPENMP)
 void SseMth1(void) {
    Int i,z;
    if (!HaveSSE()) return;
    Copy();
    for (z=0;z<Z;z++) {
        #pragma omp parallel for
        for (i=0;i<N;i++) {
            Int j;
            for (j=0;j<N;j++) Sse(i,j);
        }
        Progress("#SseMth(1) ",z,Z);
    }
    PrintSum("SSE    multi  thread(1)");
 }
 
 void SseMth2(void) {
    Int z,n=0;
    if (!HaveSSE()) return;
    Copy();
    #pragma omp parallel for
    for (z=0;z<Z;z++) {
        Int i,j;
        for (i=0;i<N;i++) for (j=0;j<N;j++) Sse(i,j);
        #pragma omp critical
        Progress("#SseMth(2) ",n++,Z);
    }
    PrintSum("SSE    multi  thread(2)");
 }
  #else
 void SseSth(void) {
    Int i,j,z;
    Copy();
    if (!HaveSSE()) return;
    for (z=0;z<Z;z++) {
        for (i=0;i<N;i++) for (j=0;j<N;j++) Sse(i,j);
        Progress("#SseSth ",z,Z);
    }
    PrintSum("SSE    single thread(1)");
 }
  #endif

  #if AVX
   #if defined(__GNUC__)
    #pragma GCC push_options
    #pragma GCC target("avx")
   #endif
 void Avx(Int i,Int j) {
    Int k; fvec left=C[i],right=T[j];
    float u[8]; __m256 m;
    m=_mm256_setzero_ps();
    for (k=0;k<N;k+=8)
      m=_mm256_add_ps(m,
          _mm256_mul_ps(*(__m256*)&left[k],*(__m256*)&right[k]));
    _mm256_storeu_ps(u,m);
    A[i][j]=(float)(u[0]+u[1]+u[2]+u[3]+u[4]+u[5]+u[6]+u[7]);
 }
   #if defined(__GNUC__)
    #pragma GCC pop_options
   #endif

   #if defined(_OPENMP)
 void AvxMth1(void) {
    Int i,z;
    if (HaveAVX()==0) return;
    Copy();
    for (z=0;z<Z;z++) {
        #pragma omp parallel for
        for (i=0;i<N;i++) {
            Int j;
            for (j=0;j<N;j++) Avx(i,j);
        }
        Progress("#AvxMth(1) ",z,Z);
    }
    PrintSum("AVX    multi  thread(1)");
 }
 void AvxMth2(void) {
    Int z,n=0;
    if (HaveAVX()==0) return;
    Copy();
    #pragma omp parallel for
    for (z=0;z<Z;z++) {
        Int i,j;
        for (i=0;i<N;i++) for (j=0;j<N;j++) Avx(i,j);
        #pragma omp critical
        Progress("#AvxMth(2) ",n++,Z);
    }
    PrintSum("AVX    multi  thread(2)");
 }
   #else
 void AvxSth(void) {
    Int i,j,z;
    if (HaveAVX()==0) return;
    Copy();
    for (z=0;z<Z;z++) {
        for (i=0;i<N;i++) for (j=0;j<N;j++) Avx(i,j);
        Progress("#AvxSth ",z,Z);
    }
    PrintSum("AVX    single thread(1)");
 }
   #endif
  #endif
 #endif

#endif
