
 /*-------------------------------------------------------------*
  |                                                             |
  | OptQ.c                                                      |
  |                                                             |
  |   QNA(Queueing Network Analyzer)test program                |
  |                                                             |
  | References                                                  |
  |   [1] QNA:Queueing Network Analyzer                         |
  |        by W. Whitt (1983)                                   |
  |   [2] OR library  Applied queueing theory                   |
  |        by Hidenori MORIMURA                                 |
  |                                                             |
  *-------------------------------------------------------------*/
#define FILE_NAME "OptQ.c"

#if defined(PDEF_H)

 #include "../anneal.h"
LIST("# " FILE_NAME " version 8.16")
PROC(Solve,EJ("Solve QNA","ＱＮＡで評価を行なう"))
INT(TN,2,  EJ("Model title line size.","データ ファイルのタイトルの行数"))
SVEC(Title, 0,TN,EJ("Model title.","データ ファイルのタイトル"))

          /*-----------------------------------*
           |  Node input variables             |
           *-----------------------------------*/
INT(N,       3,  EJ("Num of nodes.","内部ノード数"))
CDEF(NameN,  8,N,EJ("Node Name.","ノード名"))
SVEC(MsgN,  0,N,EJ("comment Message of the node.","ノードに対するコメント"))
RVEC(AvaiN,305,N,EJ("Availability of the node.","ノードの可動率"))
IVEC(SrN,    3,N,EJ("Num of servers at the node.","窓口数"))

          /*-----------------------------------*
           |  Node output variables            |
           *-----------------------------------*/
RVEC(AvArN,306,N,EJ("Average Node external arrival.",
                    "ノードに対する外部到着率"))
RVEC(CvExN,105,N,EJ("Cv2 of Node external arrival.",
                    "ノードに対する外部到着率の変動係数"))
RVEC(CvArN,105,N,EJ("Cv2 of Node arrival.",
                    "ノードに対する内部到着率の変動係数"))
RVEC(AvStN,208,N,EJ("Average Node service time.",
                    "ノードでの平均サービス時間"))
RVEC(CvStN,105,N,EJ("Cv2 of Node service time.",
                    "ノードでのサービス時間の変動係数"))
RVEC(UtilN,305,N,EJ("Utilization of node.","ノードの稼働率"))
RVEC(FlowN,306,N,EJ("Internal flow rate of node.","ノードでの内部到着率"))
RVEC(DeptN,306,N,EJ("Node external departure rate.","ノードでの退去率"))
RVEC(CustN,308,N,EJ("Num of Customers in node j.","ノードでの平均系内数"))
RVEC(QLenN,308,N,EJ("Num of waiting Customers.","ノードでの平均待ち行列長"))
RVEC(WaitN,308,N,EJ("Average Waiting time at node j.",
                    "ノードでの平均待ち時間"))

          /*-----------------------------------*
           |  Job input variables              |
           *-----------------------------------*/
INT(J,       2,        EJ("Num of Jobs.","ジョブ数"))
CDEF(NameJ,  8,J,      EJ("Job name.","ジョブ名"))
SVEC(MsgJ,   0,J,      EJ("comment Message of job.",
                          "ジョブに対するコメント"))
RVEC(AvArJ,106,J,      EJ("Average external arrival rate.",
                          "ジョブに対する外部到着率"))
RVEC(CvArJ,105,J,      EJ("Cv2 of external arrival rate.",
                          "ジョブに対する外部到着の変動係数"))
REAL(CoefJ,709,        EJ("Multiplier of all arrival rate.",
                          "ジョブに対する外部到着率の乗数(省略可)"))
IVEC(L,      3,J,      EJ("Num of nodes on Job k.","ジョブ上の工程数"))
CJAG(PathL,  8,L,NameN,EJ("Routing Nodes of Job k.","工程のノード名"))
RJAG(AvStL,106,L,      EJ("Average Service time at each.",
                          "工程の平均サービス時間"))
RJAG(CvStL,103,L,      EJ("Cv2 of Service time at each.",
                          "工程のサービス時間の変動係数"))

          /*-----------------------------------*
           |  Job output variables             |
           *-----------------------------------*/
RVEC(AvAr0,106,J,EJ("Modified external arrival rate.",
                    "乗数をかけた外部到着率"))
RVEC(RespJ,208,J,EJ("Response time of Job k.","ジョブに対応する応答時間"))

          /*-----------------------------------*
           |  Node matrix variables            |
           *-----------------------------------*/
RMAT(RoutM,306,N,N,EJ("Routing matrix.","経路行列"))
RMAT(FlowM,306,N,N,EJ("Flow rate.","流出行列"))

          /*-----------------------------------*
           |  Temporary variables              |
           *-----------------------------------*/
RMAT(GausM,306,N,N,"Work Matrix for Gauss Jordan.")
RMAT(B55,  306,N,N,"Work Matrix for eq.5-5.")
RMAT(P55,  306,N,N,"Work Matrix for eq.5-5.")
RVEC(A55,  306,N,  "Work Vector for eq.5-5.")
RVEC(N55,  306,N,  "Work Vector for eq.5-5.")

          /*-----------------------------------*
           |  Variables for optimization       |
           *-----------------------------------*/
IVEC(SrMin,     5,N,EJ("Anneal min num of servers.",
                       "アニーリングの探索範囲"))
IVEC(SrMax,     5,N,EJ("Anneal max num of servers.",
                       "アニーリングの探索範囲"))
RVEC(TrgQN,   306,N,EJ("Anneal target of QLenN.","目標とする中間在庫量"))
REAL(TrgCoef, 307,  EJ("modify TrgQN.","TrgQN を修正する"))
RVEC(SrCost,  306,N,EJ("Cost of servers.","設備コストに対する罰金"))
RVEC(QLenCost,306,N,EJ("Cost of queue length.","中間在庫に対する罰金"))
RVEC(RespCost,306,J,EJ("Cost of response time.","反応時間に対する罰金"))
INT(Saturate,   8,EJ("System saturate flag.","ノードがパンクした"))
FUNC(QueDiff,  EJ("Object Function (diff of queue-len)",
                  "中間在庫を目標に近づけるための評価関数"))
FUNC(TotalCost,EJ("Cost Function (RespJ+SrX+QLen)",
                  "コストを考慮した評価関数(RespJ+SrN+QLen)"))

AUTO("CoefJ TrgCoef")
AUTO("1     1")
AUTO("Range J > 0")
AUTO("Range L > 0")
AUTO("Range N >= 0")
AUTO("Range SrN > 0")
AUTO("Range AvArN > 0")
AUTO("Range AvStN > 0")
AUTO("Range AvArJ > 0")
AUTO("Range CoefJ > 0")
AUTO("Range AvStL > 0")
AUTO("Range SrMin > 0")
AUTO("Range SrMax > 0")
AUTO("Range AvaiN > 0 <= 1")
AUTO("Range CvExN >= 0")
AUTO("Range CvArN >= 0")
AUTO("Range CvStN >= 0")
AUTO("Range CvArJ >= 0")
AUTO("Range CvStL >= 0")
AUTO("Range TrgCoef >= 0")

#else

 #if defined(__TURBOC__) && !defined(__BORLANDC__)
  #define NO_ADD        1
  #define NO_SORT       1
  #define NO_UTF16      1
 #endif
 #if defined(LSI_C)
  #define NO_VERSION    1
  #define NO_ADD        1
  #define ASCII         1
  #define NO_ALIAS      1
  #define NO_APPEND     1
  #define NO_CDEF       0
  #define NO_CLOCK      1
  #define NO_CSV        1
  #define NO_TSV        1
  #define NO_ECHO       0
  #define NO_BOM        1
  #define NO_CHECK_CDEF 1
  #define NO_FAST_STR   1
  #define NO_FORMAT     1
  #define NO_FREE       1
  #define NO_JAG        0
  #define NO_LAP        1
  #define NO_LIMIT      1
  #define NO_LIST       1
  #define NO_NINT       1
  #define NO_HANGUL     1
  #define NO_OUTPUT     0
  #define NO_PROGRESS   1
  #define NO_RAND       0
  #define NO_RANGE      0
  #define NO_SAVE       1
  #define NO_SORT       1
  #define NO_STR        0
  #define NO_SYSTEM     1
  #define NO_DELETE     1
  #define NO_UTF16      1
  #define NO_VMJ        1
 #endif
 #include "pocom.h"
CHECK_FILE_NAME

 #if FLOAT_REAL
 static real FactTable[35];
 #else
static real FactTable[70];
 #endif


void GaussJordan(rmat Q,rvec lm0,rvec lm) {
    Int i,j,k; real pivot;
    for (i=0;i<N;i++) {
        for (j=0;j<N;j++) if (i==j) GausM[i][j]=1-Q[j][i];
                           else     GausM[i][j]=  - Q[j][i];
        lm[i]=lm0[i];
    }
    for (k=0;k<N;k++) {
        pivot=GausM[k][k];
        for (j=k+1;j<N;j++) GausM[k][j]/=pivot;
        lm[k]/=pivot;
        for (i=0;i<N;i++) if (i!=k) {
            pivot=GausM[i][k];
            for (j=k+1;j<N;j++) GausM[i][j]-=pivot*GausM[k][j];
            lm[i]-=pivot*lm[k];
        }
    }
}

void InitQna(void) {
    Check("AvaiN CvArJ CvStL NameN SrN NameJ AvArJ L PathL AvStL");
    Alloc("AvArN CvExN CvArN AvStN CvStN UtilN");
    Alloc("FlowN DeptN CustN QLenN WaitN AvAr0");
    Alloc("RespJ RoutM FlowM GausM B55 P55 A55 N55");
}

void SolveInterFlow(void) {
    Int i,j,k,p,nk,*P;
    real lk,Slik,service,variance,ArrSum,SerSum,Cv2Sum;
    for (k=0;k<J;k++) {
        nk=L[k];
        P=PathL[k];
        AvAr0[k]=lk=AvArJ[k] * CoefJ;
        AvArN[P[0]]   +=lk;
        DeptN[P[nk-1]]+=lk;
        for (p=0;p<nk-1;p++)  FlowM[P[p]][P[p+1]]+=lk;
    }
    for (i=0;i<N;i++) {
        Slik=DeptN[i];
        for (j=0;j<N;j++) Slik+=FlowM[i][j];
        if (Slik>0.0) for (j=0;j<N;j++)
            RoutM[i][j]=FlowM[i][j]/Slik;/* (2.4) */
        ArrSum=0.0;  SerSum=0.0;  Cv2Sum=0.0;
        for (k=0;k<J;k++) {
            lk=AvAr0[k];
            for (p=0,nk=L[k]; p<nk; p++) if (PathL[k][p]==i) {
                service =AvStL[k][p]/AvaiN[i];
                variance=CvStL[k][p]+1;
                ArrSum+=lk;
                SerSum+=lk*service;
                Cv2Sum+=lk*service*service*variance;
            }
        }
        if (ArrSum>0.0) AvStN[i]=SerSum/ArrSum; /* (2.5) */
        if (SerSum>0.0)
            CvStN[i]=Cv2Sum*ArrSum/(SerSum*SerSum)-1; /* (2.6) */
    }
    GaussJordan(RoutM,AvArN,FlowN);
}

void CheckUtil(void) {
    Int i; bool ErrorFlag=false; real Utilization,p=0.99;
    for (i=0;i<N;i++) {
        Utilization=FlowN[i]*AvStN[i]/(real)SrN[i];  /* (2.6) */
        UtilN[i]   =Utilization;
        if (Utilization>p) ErrorFlag=true;
    }
    if (ErrorFlag) Saturate=1; else Saturate=0;
}

void PutTrafErr(void) {
    Int n; static char buf[60]; size_t pos=0,i; char*p;
    for (n=0;n<N;n++) if (UtilN[n]>0.99) {
        i=sizeof buf-1-pos;
        p=Left__(NameN[n],(int)i);
        for (i=0;p[i]&&pos<sizeof buf-1;i++,pos++) buf[pos]=p[i];
        if (pos<sizeof buf-1) buf[pos++]=' '; else break;
    }
    buf[pos]=0;
    Error(buf,EJ("Node saturated","ノードがパンクした"));
}

void SolveCvExN(void) {
    Int i,k; real vi,wi,pi,l0i,Sck2lk,t;
    for (i=0;i<N;i++)
     if (AvArN[i]==0.0) CvExN[i]=1.0; 
      else {
          pi =UtilN[i];
          l0i=AvArN[i];
          vi =0.0;
          for (k=0; k<J; k++) if (PathL[k][0]==i)
          { t=AvAr0[k]/l0i; vi+=t*t; }
          vi=1/vi;                                   /* (2.9) */
          wi=1/(1+4*(1-pi)*(1-pi)*(vi-1)); /* (2.8) */
          Sck2lk=0;
          for (k=0;k<J;k++) if (PathL[k][0]==i)
              Sck2lk+=CvArJ[k]*AvAr0[k];
          CvExN[i]=(1-wi)+wi*Sck2lk/l0i;             /* (2.7) */
      }
}

void Solve5_5(void) {
    real SumA,vj,wj,qij,pij,p0j,ui,ui2,pj,maxim; Int i,j;
    for (j=0;j<N;j++) {
        if (FlowN[j]!=0.0) for (i=0;i<N;i++)
            P55[i][j]=FlowM[i][j]/FlowN[j];
        maxim=CvStN[j];
        if (maxim<0.2) maxim=0.2;
        N55[j]=1+(maxim-1)/sqrt((double)SrN[j]); /* (5.8) */
    }
    for (j=0;j<N;j++) {
        if (FlowN[j]!=0.0) p0j=AvArN[j]/FlowN[j]; else p0j=0.0;
        for (i=0,vj=p0j*p0j;i<N;i++) { pij=P55[i][j]; vj+=pij*pij; }
        if (vj!=0) vj=1/vj; else vj=1;         /* (5.11) */
        pj=UtilN[j];
        wj=1/(1+4*(1-pj)*(1-pj)*(vj-1));/* (5.10) */
        for (i=0,SumA=0.0;i<N;i++) {
            ui=UtilN[i];
            ui2=ui*ui;
            pij=P55[i][j];
            qij=RoutM[i][j];
            SumA+=pij*((1-qij)+qij*ui2*N55[i]);
            B55[i][j]=wj*pij*qij*(1-ui2);
        }
        A55[j]=1+wj*(p0j*CvExN[j]-1+SumA);
    }
    GaussJordan(B55,A55,CvArN);
}

void SolveRespJ(void) {
    Int i,k,nk; real l,Wq,sum;
    for (i=0;i<N;i++) {
        Wq=WaitN[i];
        l=FlowN[i];
        CustN[i]=l*Wq+l*AvStN[i];
        QLenN[i]=l*Wq;
    }
    for (k=0;k<J;k++) {
        for (i=0,nk=L[k],sum=0.0;i<nk;i++)
            sum+=WaitN[PathL[k][i]]+AvStL[k][i]/AvaiN[PathL[k][i]];
        RespJ[k]=sum;
    }
}

void SolveExact(void) {
    Int s,i,n; real p0,l,t,u,a,ds,an,as,s_1f,su_l,Sann;
    if (FactTable[0]==0.0) {
        FactTable[0]=1.0;
        for (i=1;i<(Int)(sizeof(FactTable)/sizeof(real));i++)
            FactTable[i]=FactTable[i-1]*(real)i;
    }
    SolveCvExN();
    Solve5_5();
    for (i=0;i<N;i++) {
        s=SrN[i];
        if (s>=(Int)(sizeof(FactTable)/sizeof(real))) WaitN[i]=0.0;
         else {
            l=FlowN[i];
            t=AvStN[i];
            if (t>0) u=1/t; else u=1;
            a=l*t;
            s_1f=FactTable[s-1];
            ds=s;
            su_l=ds*u-l;
            for (n=0,Sann=0.0,an=1.0;n<s;n++)
            { Sann+=an/FactTable[n]; an*=a; }
            as=an;
            p0=1/(Sann+as/(s_1f*(ds-a)));
            WaitN[i]=  (CvArN[i]+CvStN[i])*0.5*p0*u*as
                      /(s_1f*su_l*su_l);
         } 
    }
    SolveRespJ();
}

void SolveApprox(void) {
    Int i; real s,x,m,p=0.99;
    for (i=0;i<N;i++) {
        x=UtilN[i];
        s=AvStN[i];
        m=SrN[i];
        if (SrN[i]>=(Int)(sizeof(FactTable)/sizeof(real))) WaitN[i]=0.0;
         else if (x<=p) WaitN[i]=(1+CvStN[i])*0.5*x*s/(1-x)/m;
         else   WaitN[i]=(1+CvStN[i])*0.5*s*(p+(x-p)
              / (1-p))/(1-p)/m;
    }
    SolveRespJ();
}

void Solve(void) {
    InitQna(); SolveInterFlow(); CheckUtil();
    if (Saturate) PutTrafErr(); else SolveExact();
}

 /*-------------------------------------------------------------*
  |                                                             |
  |       ANNEALING APPLICATION DEPENDENT ROUTINES              |
  |                                                             |
  *-------------------------------------------------------------*/
void TotalCost(const char*Argument) {
    Int i;   real Cost=0.0;
    if (Argument==0||Count==0.0) {
        Check("SrCost");
        if (RowSize(QLenCost)) Check("QLenCost");
        if (RowSize(RespCost)) Check("RespCost");
        InitQna();
        SolveInterFlow();
    }
    CheckUtil();
    if (Saturate) SolveApprox(); else SolveExact();
    if (RowSize(SrCost))   for (i=0;i<N;i++) Cost+=SrN[i]*SrCost[i];
    if (RowSize(QLenCost)) for (i=0;i<N;i++) Cost+=QLenN[i]*QLenCost[i];
    if (RowSize(RespCost)) for (i=0;i<J;i++) Cost+=RespJ[i]*RespCost[i];
    CurrScore=Cost;
}

void QueDiff(const char*Argument) {
    Int i; real SumOfDifference=0.0;
    if (Argument==0||Count==0.0) {
        Check("TrgQN"); InitQna();
        SolveInterFlow();
    }
    CheckUtil();
    if (Saturate) SolveApprox(); else SolveExact();
    for (i=0;i<N;i++)
        SumOfDifference+=fabs(QLenN[i]-TrgQN[i]*TrgCoef);
    CurrScore=SumOfDifference;
}

#endif
