#define FILE_NAME "Stat.c"

#if defined(PDEF_H)

LIST("# " FILE_NAME " version 8.16")
INT(N,10,           EJ("Num of data","個体の数"))
INT(M,10,           EJ("Num of variable","変数の数"))
INT(P,10,           EJ("Num of principal component","主成分の数"))
INT(MaxPrin,7,      EJ("Max of principal component","主成分の最大数"))
REAL(QrEps,-309,    EJ("QR alg.'s allowed error","QR法の許容誤差"))
REAL(QrMaxIter,9,   EJ("QR alg.'s max iter.","QR法の最大繰返し数"))
REAL(LsqEps,-309,   EJ("Lsq's allowed error","重回帰分析の許容誤差"))

CDEF(Var,    10,M,  EJ("Name of variable","説明変量名"))
SVEC(Name,   10,N,  EJ("Name of data","個体名"))
SVEC(Prin,   10,P,  EJ("Name of principal component","主成分名"))
RMAT(Data,  510,M,N,EJ("Input data","入力データ"))
RMAT(X,     510,M,N,EJ("Intermediate data","中間データ"))
RVEC(Mean,  510,M,  EJ("Mean ","平均値"))
RVEC(StdDev,510,M,  EJ("Standard deviation","標準偏差"))
RMAT(Q,     510,M,M,EJ("Transformation matrix","変換行列"))
RMAT(Corr,  205,M,M,EJ("Corr. Coef. matrix","相関係数行列"))
RVEC(Work,  510,M,  EJ("Work vector","作業用"))
RVEC(Lambda,510,M,  EJ("Eigen value","固有値"))
RVEC(Lamb,  510,P,  EJ("Eigen value(2)","固有値2"))
RVEC(Prop,  510,P,  EJ("Proportion","寄与率"))
RVEC(CProp, 510,P,  EJ("Cumulation proportopm","累積寄与率"))
RMAT(Coef,  510,M,P,EJ("Coefficient","係数"))
RMAT(Score, 510,N,P,EJ("Score of principal component","主成分得点"))

INT(R,2,       EJ("Rank","ランク"))
RVEC(T,206,P,  EJ("t value","t値"))
RVEC(B,509,P,  EJ("Regression coefficient","回帰係数"))
RVEC(S,206,P,  EJ("Standard error","標準誤差"))
RVEC(Normsq,206,P,EJ("Sum of squares","２乗和"))
REAL(Rss,306,EJ("residual sum of squares","残差２乗和"))
INT(Phi,3,EJ("Degree of freedom","自由度"))
REAL(Mse,306,EJ("Mean square of error","平均２乗残差"))
REAL(CorrCoef,307,EJ("Corr. Coef.","相関係数"))
RVEC(InitNormsq,206,P,EJ("Initial norm","初期ノルム"))
CVEC(Col,10,P,Var,EJ("Explanatory variable","説明変量"))

RMAT(A,307,M,M,EJ("Coef. matrix","係数行列"))
RVEC(D,307,M,  EJ("Right side vector","右辺"))
RMAT(C,307,M,M,EJ("Work matrix","作業用の行列"))
RVEC(Y,509,M,  EJ("Multiple regression equation","重回帰式"))
IVEC(Ip,2,M,EJ("Row swapping info.","行交換情報"))
RVEC(Weight,206,M,EJ("Reciprocal of absolute max vaule","最大絶対値の逆数"))
REAL(Det,206,EJ("Determinant","行列式"))
RVEC(Forecast,206,N,EJ("Forecast","推測値"))
REAL(SOrg,206,EJ("Variance of actiual","実測値の分散"))
REAL(SFor,206,EJ("Variance of forecast","推測値の分散"))
REAL(R2,307,EJ("Proportion","寄与率"))
REAL(R1,307,EJ("multiple corr.","重相関係数"))

PROC(Lu,EJ("LU decomposition","ＬＵ分解"))
PROC(Sol,EJ("Solve equation","ＬＵ分解の結果を使って方程式を解く"))
PROC(Principal,EJ("Principal component analysis","主成分分析の実行"))
PROC(Regress1,EJ("Multiple regression analysis(1:No Pivot)",
     "重回帰分析の実行その１(No Pivot)"))
PROC(Regress2,EJ("Multiple regression analysis(2:No Pivot)",
     "重回帰分析の実行その２(Pivoting)"))
PROC(Regress3,EJ("Multiple regression analysis(3)","重回帰分析の実行その3"))
PROC(MakeData,EJ("Make random data","データを乱数で作成"))
PROC(Standard,EJ("Standardization","dataを標準化する"))

AUTO("QrEps LsqEps QrMaxIter MaxPrin")
AUTO(" 1e-6   1e-6       100       5")
AUTO("Range M > 0")
AUTO("Range CorrCoef >= -1 <= 1")

#else

 #if defined(LSI_C)
  #define NO_AFTER_VER2 1
 #endif

#include "pocom.h"
CHECK_FILE_NAME

/* 内積 */
real innerproduct(Int n,rvec x,rvec y) {
    Int i; real sum=0;
    for (i=0;i<n;i++) sum+=x[i]*y[i];
    return sum;
}

/* 平方和 */
real sumofsquares(Int n,rvec x) {
    Int i; real sum=0;
    for (i=0;i<n;i++) sum+=x[i]*x[i];
    return sum;
}

/* ハウスホルダー変換 */
real house(Int n,rvec x) {
    Int i; real s,t;
    s=sqrt(sumofsquares(n,x));
    if (s!=0) {
        if (x[0]<0) s=-s;
        x[0]+=s;
        t=1.0/sqrt(x[0]*s);
        for (i=0;i<n;i++) x[i]*=t;
    }
    return -s;
}

/* ３重対角化 */
void tridiagonalize(Int n,rmat a,rvec d,rvec e) {
    Int i,j,k; real s,t,p,q; rvec v,w;
    for (k=0;k<n-2;k++) {
        v=a[k]; d[k]=v[k]; e[k]=house(n-k-1,&v[k+1]);
        if (e[k]==0) continue;
        for (i=k+1;i<n;i++) {
            s=0;
            for (j=k+1;j<i;j++) s+=a[j][i]*v[j];
            for (j=i;  j<n;j++) s+=a[i][j]*v[j];
            d[i]=s;
        }
        t=innerproduct(n-k-1,&v[k+1],&d[k+1])/2.0;
        for (i=n-1;i>k;i--) {
            p=v[i]; q=d[i]-t*p; d[i]=q;
            for (j=i;j<n;j++) a[i][j]-=p*d[j]+q*v[j];
        }
    }
    if (n>=2) { d[n-2]=a[n-2][n-2]; e[n-2]=a[n-2][n-1]; }
    if (n>=1)   d[n-1]=a[n-1][n-1];
    if (n>=1) {
        v=a[n-1];
        for (i=0;i<n-1;i++) v[i]=0;
        v[n-1]=1;
    }
    if (n>=2) {
        v=a[n-2];
        for (i=0;i<n;i++) v[i]=0;
        v[n-2]=1;
    }
    for (k=n-3;k>=0;k--) {
        v=a[k];
        for (i=k+1;i<n;i++) {
            w=a[i];
            t=innerproduct(n-k-1,&v[k+1],&w[k+1]);
            for (j=k+1;j<n;j++) w[j]-=t*v[j];
        }
        for (i=0;i<n;i++) v[i]=0;
        v[k]=1;
    }
}

/* QR 法 */
void eigen(Int n,rmat a,rvec d,rvec e) {
    Int i,j,k,h,iter; real c,s,t,w,x,y;
    tridiagonalize(n,a,d,&e[1]); e[0]=0;
    for (h=n-1;h>0;h--) {
        j=h;
        while (j&&fabs(e[j])>QrEps*(fabs(d[j-1])+fabs(d[j]))) j--;
        if (j==h) continue;
        iter=0;
        do {
            if (++iter>QrMaxIter)
                Error("eigen",EJ("Cannot converge","収束しませんでした"));
            w=(d[h-1]-d[h])/2.0; t=e[h]*e[h]; s=sqrt(w*w+t);
            if (w<0) s=-s;
            x=d[j]-d[h]+t/(w+s); y=e[j+1];
            for (k=j;k<h;k++) {
                if (fabs(x)>=fabs(y))
                     { t=-y/x; c=1.0/sqrt(t*t+1.0); s=t*c; }
                else { t=-x/y; s=1.0/sqrt(t*t+1.0); c=t*s; }
                w=d[k]-d[k+1]; t=(w*s+2.0*c*e[k+1])*s;
                d[k]-=t; d[k+1]+=t;
                if (k>j) e[k]=c*e[k]-s*y;
                e[k+1]+=s*(c*w-2.0*s*e[k+1]);
                for (i=0;i<n;i++) {
                    x=a[k][i]; y=a[k+1][i];
                    a[k][i]=c*x-s*y;
                    a[k+1][i]=s*x+c*y;
                }
                if (k<h-1) {
                    x=e[k+1]; y=-s*e[k+2];
                    e[k+2]*=c;
                }
            }
        } while (fabs(e[h])>QrEps*(fabs(d[h-1])+fabs(d[h])));
    }
    for (k=0;k<n-1;k++) {
        h=k; t=d[h];
        for (i=k+1;i<n;i++)
            if (d[i]>t) { h=i; t=d[h]; }
        d[h]=d[k]; d[k]=t;
        SwapRow(a,h,k);
    }
}

void Principal(void) {
    Int i,j,k; real s,t,u,ndf; char buf[40];
    Check("Data"); ndf=N-1.0;
    if (M>MaxPrin) P=MaxPrin; else P=M;
    Alloc("X Q Lambda Work Coef Score Prin");
    Alloc("Mean StdDev Lamb Prop CProp Corr"); 
    for (j=0;j<M;j++) for (i=0;i<N;i++) X[j][i]=Data[j][i];
    for (i=0;i<P;i++) {
        sprintf(buf,EJ("%" PRIz "d-comp","第%" PRIz "d主成分"),i+1);
        SetSvec(Prin,i,buf);
    }
    for (j=0;j<M;j++) {
        t=0;
        for (i=0;i<N;i++) t+=X[j][i];
        t/=N;
        for (i=0;i<N;i++) X[j][i]-=t;
        u=sumofsquares(N,X[j])/ndf; s=sqrt(u);
        Mean[j]=t; StdDev[j]=s; Corr[j][j]=1;
        for (i=0;i<N;i++) X[j][i]/=s;
    }
    for (j=0;j<M;j++) for (k=0;k<j;k++)
        Corr[j][k]=Corr[k][j]=innerproduct(N,X[j],X[k])/ndf;
    for (j=0;j<M;j++) for (i=0;i<M;i++) Q[i][j]=Corr[i][j];
    eigen(M,Q,Lambda,Work);
    t=0;
    for (k=0;k<M;k++) t+=Lambda[k];
    s=0;
    for (k=0;k<P;k++) {
        Lamb[k]=Lambda[k];
        Prop[k]=Lamb[k]/t;
        s+=Prop[k];
        CProp[k]=s;
    }
    for (j=0;j<M;j++) for (k=0;k<P;k++) Coef[j][k]=Q[k][j];
    for (i=0;i<N;i++) for (k=0;k<P;k++) {
        s=0;
        for (j=0;j<M;j++) s+=Q[k][j]*X[j][i];
        Score[i][k]=s;
    }
}

void invr(void) {
    Int i,j,k; real s;
    for (k=0;k<R;k++) {
        X[k][k]=1/X[k][k];
        for (j=k-1;j>=0;j--) {
            s=0;
            for (i=j+1;i<=k;i++) s-=X[i][j]*X[i][k];
            X[j][k]=s*X[j][j];
        }
    }
}

void disp(void) {
    Int j;
    Rss=sumofsquares(N-R,&X[P][R]);
    invr();
    for (j=0;j<R;j++) {
        S[j]=sqrt(sumofsquares(R-j,&X[j][j])*Rss/((real)N-R));
        if (S[j]>0) T[j]=fabs(B[j]/S[j]);
    }
    Phi=N-R; Mse=sqrt(Rss/Phi);
}

void Regress1(void) {
    Int i,j,k; real s,t,u; rvec v,w;
    Check("Data"); P=M-1; Alloc("Col X B S T Normsq");
    for (i=0;i<N;i++) for (j=0;j<M;j++) X[j][i]=Data[j][i];
    for (j=0;j<P;j++) { Normsq[j]=sumofsquares(N,X[j]); Col[j]=j; }
    for (R=k=0;k<P;k++) {
        if (Normsq[k]==0) continue;
        v=X[k]; u=sumofsquares(N-R,&v[R]);
        if (u/Normsq[k]<LsqEps*LsqEps) continue;
        /*X[R]=v;*/ u=sqrt(u);
        if (v[R]<0) u= -u;
        v[R]+=u; t=1/(v[R]*u);
        for (j=k+1;j<=P;j++) {
            w=X[j];
            s=t*innerproduct(N-R,&v[R],&w[R]);
            for (i=R;i<N;i++) w[i]-=s*v[i];
        }
        v[R]= -u; R++;
    }
    for (j=R-1;j>=0;j--) {
        s=X[P][j];
        for (i=j+1;i<R;i++) s-=X[i][j]*B[i];
        B[j]=s/X[j][j];
    }
    disp();
}

void Regress2(void) {
    Int i,j; real s,t,u; rvec v,w;
    Check("Data"); P=M-1;
    Alloc("X Col B S T Normsq InitNormsq");
    for (i=0;i<N;i++) for (j=0;j<M;j++) X[j][i]=Data[j][i];
    for (j=0;j<P;j++) {
        Col[j]=j;
        Normsq[j]=sumofsquares(N,X[j]);
        InitNormsq[j]=Normsq[j]!=0?Normsq[j]:-1;
    }
    for (R=0;R<P;R++) {
        if (R!=0) {
            j=R; u=0;
            for (i=R;i<P;i++) {
                t=Normsq[i]/InitNormsq[i];
                if (t>u) { u=t; j=i; }
            }
            SwapRow(Col,j,R);
            SwapRow(Normsq,j,R);
            SwapRow(InitNormsq,j,R);
            SwapRow(X,j,R);
        }
        v=X[R]; u=sumofsquares(N-R,&v[R]);
        if (u/InitNormsq[R]<LsqEps*LsqEps) break;
        u=sqrt(u); if (v[R]<0) u= -u;
        v[R]+=u; t=1/(v[R]*u);
        for (j=R+1;j<=P;j++) {
            w=X[j];
            s=t*innerproduct(N-R,&v[R],&w[R]);
            for (i=R;i<N;i++) w[i]-=s*v[i];
            if (j<P) Normsq[j]-=w[R]*w[R];
        }
        v[R]= -u;
    }
    for (j=R-1;j>=0;j--) {
        s=X[P][j];
        for (i=j+1;i<R;i++) s-=X[i][j]*B[i];
        B[j]=s/X[j][j];
    }
    disp();
}

void MakeData(void) {
    Int i,j; real r1,r2,s;
    Check("M"); Alloc("Data");
    if (RowSize(Var)==0) Alloc("Var");
    for (i=0;i<N;i++) {
        Data[0][i]=1;
        for (j=1;j<M;j+=2) {
            do {
                r1=Unif()*2-1;
                r2=Unif()*2-1;
                s=r1*r1+r2*r2;
            } while (s>1||s==0);
            s= -log(s)/s; r1=sqrt((1+CorrCoef)*s)*r1;
            r2=sqrt((1-CorrCoef)*s)*r2; Data[j][i]=r1+r2;
            if (j+1<M) Data[j+1][i]=r1-r2;
        }
    }
}

void Standard(void) {
    Int i,j; real sum,mean,sd,diff;
    Check("Data");
    for (j=0;j<M;j++) {
        sum=0;
        for (i=0;i<N;i++) sum+=Data[j][i];
        mean=sum/N;
        sum=0;
        for (i=0;i<N;i++) {
            diff=Data[j][i]-mean;
            sum+=diff*diff;
        }
        sd=sqrt(sum/(N-1.0));
        if (sd==0.0) Error("Standard",EJ("Denominator is 0","分母が0"));
        Data[j][i]=(Data[j][i]-mean)/sd;
    }
}

void Lu(void) {
    Int i,j,k,ii,ik; real t,u;
    Check("A"); Alloc("C Weight Ip");
    for (i=0;i<M;i++) for (j=0;j<M;j++) C[i][j]=A[i][j];
    for (k=0;k<M;k++) {
        Ip[k]=k; u=0;
        for (j=0;j<M;j++) { t=fabs(C[k][j]); if (t>u) u=t; }
        if (u==0) Error("Lu",EJ("Cannot solved","解けません"));
        Weight[k]=1/u;
    }
    Det=1.0; j=0;
    for (k=0;k<M;k++) {
        u=-1;
        for (i=k;i<M;i++) {
            ii=Ip[i]; t=fabs(C[ii][k])*Weight[ii];
            if (t>u) { u=t; j=i; }
        }
        ik=Ip[j];
        if (j!=k) { Ip[j]=Ip[k]; Ip[k]=ik; Det=-Det; }
        u=C[ik][k]; Det*=u;
        if (u==0) Error("Lu",EJ("Cannot solved","解けません"));
        for (i=k+1;i<M;i++) {
            ii=Ip[i]; t=(C[ii][k]/=u);
            for (j=k+1;j<M;j++) C[ii][j]-=t*C[ik][j];
        }
    }
}

void Sol(void) {
    Int i,j,ii; real t;
    Check("C Ip D"); Alloc("Y");
    for (i=0;i<M;i++) {
        ii=Ip[i]; t=D[ii];
        for (j=0;j<i;j++) t-=C[ii][j]*Y[j];
        Y[i]=t;
    }
    for (i=M-1;i>=0;i--) {
        t=Y[i]; ii=Ip[i];
        for (j=i+1;j<M;j++) t-=C[ii][j]*Y[j];
        Y[i]=t/C[ii][i];
    }
}

void Regress3(void) {
    Int i,j; real sum;
    Check("Data"); Alloc("X Mean A D Forecast");
 #if defined(STATIC) || defined(SEMISTATIC)
    if (M<1) return;
    /* コード分析での警告抑止。
       AUTO("Range M > 0") と Check("Data") で M>0 のはず */
 #endif
    for (i=0;i<M;i++) {
        sum=0;
        for (j=0;j<N;j++) sum+=Data[i][j];
        Mean[i]=sum/N;
    }
    for (i=0;i<M;i++) for (j=0;j<N;j++) X[i][j]=Data[i][j]-Mean[i];
    for (i=0;i<M;i++) for (j=0;j<=i;j++)
        A[i][j]=A[j][i]=innerproduct(N,X[i],X[j]);
    for (i=0;i<M;i++) A[M-1][i]=Mean[i];
    for (i=0;i<M;i++) { D[i]=A[i][M-1]; A[i][M-1]=0; }
    A[M-1][M-1]=1.0;
    Lu();
    Sol();
    for (i=0;i<N;i++) {
        Forecast[i]=Y[M-1];
        for (j=0;j<M-1;j++) Forecast[i]+=Y[j]*Data[j][i];
    }
    sum=SOrg=SFor=0;
    for (i=0;i<N;i++) sum+=Data[M-1][i];
    sum/=N;
    for (i=0;i<N;i++) SOrg+=(Data[M-1][i]-sum)*(Data[M-1][i]);
    sum=0;
    for (i=0;i<N;i++) sum+=Forecast[i];
    sum/=N;
    for (i=0;i<N;i++) SFor+=(Forecast[i]-sum)*(Forecast[i]);
    SOrg/=(N-1.0); SFor/=(N-1.0); R2=SFor/SOrg; R1=sqrt(R2);
}

#endif
