/*+----------------------------------------------------+
  |FILENAME : anneal.h                                 |
  |ANNEAL Version 4.05                                 |
  |Copyright (c) 2019 by 和田維作 All Rights Reserved. |
  +----------------------------------------------------+*/
#if !defined(POCOM_CODE)
 #if !NO_VERSION
LIST("# anneal.h version 5.02")
 #endif
INT(Sample,       6,EJ("Annealing initial sample",
                       "初期温度設定のためのサンプル数"))
INT(LoopSize,     8,EJ("Inner Loop Iteration.",
                       "メトロポリスの繰返し数"))
INT(MinLoop,      8,EJ("Min inner Loop Iteration.",
                       "メトロポリスの繰返し数の最小値"))
INT(DispBest,     8,EJ("Display Best immediately.",
                       "１ならば、最良値をすぐに出力する"))
INT(Accept,       6,EJ("Annealing Accept Flag",
                       "アクセプトされれば１になる"))
INT(CnfSize,      7,EJ("Size of configuration",
                       "状態の大きさ"))
IVEC(BestCnf,7,CnfSize,EJ("Best configuration",
                          "最良の状態"))
STR(EpsiList,     0,EJ("Output list for each epsilon.",
                       "評価関数計算毎に出力する変数のリスト"))
STR(InitList,     0,EJ("Output list for InitAnneal.",
                       "初期温度設定時に出力する統計量のリスト"))
STR(MetList,      0,EJ("Output list for Metropolis loop.",
                       "メトロポリス時に出力する統計量のリスト"))
REAL(Temp,     -108,EJ("Temparature.",
                       "温度(コントロール パラメータ)"))
REAL(MaxTe,    -108,EJ("End of linear area",
                       "温度を下げかたを変更する境界"))
REAL(Crit,      709,EJ("Annealing criteria",
                       "メトロポリスの標準偏差による終了条件"))
REAL(Gamma,     308,EJ("change in distribution",
                       "温度の下げる刻み(平衡分布の変化量)"))
REAL(Smooth,    308,EJ("smooth ratio of Sigma",
                       "分散の推定を滑らかにする"))
REAL(Multi,     308,EJ("multiplier for inner loop",
                       "メトロポリスの長さの乗数"))
REAL(Beta,     -108,EJ("Selection Probability",
                       "特定のムーブが起きる確率"))
REAL(Count,       8,EJ("Number of Evaluation.",
                       "評価を行なったの回数"))
REAL(MaxCount,    8,EJ("Max num of Evaluation.",
                       "評価関数計算回数の最大値"))
REAL(MaxLoop,     8,EJ("Max inner Loop Iteration.",
                       "メトロポリスの繰返し数の最大値"))
REAL(OptiScore,-209,EJ("Optimum Score",
                       "望ましい評価関数値"))
REAL(CurrScore,-209,EJ("Current Score",
                       "現在の評価関数値"))
REAL(BestScore,-209,EJ("Best Score",
                       "最良の評価関数値"))
REAL(SpaceSize,-209,EJ("Space Size.",
                       "解空間の広さ"))
REAL(GlbAcc99, -108,EJ("Initial Global Acc.",
                       "初期(温度=∞)の空間全体のエントロピー"))
REAL(LocAcc99, -108,EJ("Initial Local Acc.",
                       "初期ムーブが到達できる空間のエントロピー"))
REAL(Energy99, -108,EJ("Initial Mean Score.",
                       "空間全体での評価関数の平均値"))
REAL(Sigma99,  -108,EJ("Initial S.D.",
                       "空間全体での評価関数の標準偏差"))
REAL(Sigma,    -108,EJ("Standard Deviation.",
                       "メトロポリスでの評価関数の標準偏差"))
REAL(GlbAcc,   -108,EJ("Global Accessibility.",
                       "メトロポリスでの平衡分布のエントロピー"))
REAL(LocAcc,   -108,EJ("Local Accessibility.",
                       "メトロポリスでの評価関数のエントロピー"))
REAL(Energy,   -108,EJ("Mean Score.",
                       "メトロポリスでの評価関数の平均値"))
INT(ImpSamp,      6,EJ("Improve initial sample",
                       "繰り返し改善の初期サンプルの大きさ"))
REAL(ImpLoop,     8,EJ("Num of Imp loop size",
                       "繰り返し改善の繰り返し数"))
FUNC(Anneal,   EJ("Execute Annealing",
                   "アニーリングを実行する"))
FUNC(Improve,  EJ("Execute Iterative Improvement",
                   "繰り返し改善法を実行する"))
FUNC(PartiMove,EJ("Partition Move",
                   "集合の分割を変えるムーブ"))
FUNC(WalkMove, EJ("Multi dimension random walk move",
                   "確率1/2で整数値を１つだけ増減するムーブ"))
FUNC(PermuMove,EJ("Permutation move",
                   "集合の並べ換えをするムーブ"))

AUTO("Range Sample   >= 2")
AUTO("Range MinLoop  >= 0")
AUTO("Range DispBest >= 0 <= 1")
AUTO("Range Crit     >  0")
AUTO("Range Gamma    >= 0 <= 1")
AUTO("Range Smooth   >= 0 <= 1")
AUTO("Range Multi    >= 0")
AUTO("Range MaxCount >= 1")
AUTO("Range MaxLoop  >= 1")
AUTO("Range ImpSamp  >= 1")
AUTO("Range ImpLoop  >= 1")
AUTO("Sample MinLoop Crit Gamma Smooth Multi")
AUTO("   384      10 0.01  0.10   0.20 30.00")
AUTO("MaxCount MaxLoop OptiScore ImpSamp ImpLoop")
AUTO("10000000   10000     -1e37    1000   10000")

#else

char CalcCurrScore__[1],Generate__[1],Move__[1],UndoMove__[1];
char AcceptBest__[1],InitMove__[1],LoadBest__[1],DisplayBest__[1];
wptr EpsiArg__,MoveArg__;

real stirling(Int n) {
    real x=n; static real logpi;
    if (x==0.0) x=0.001;
    if (logpi==0) logpi=log(2.0*3.14159265358979323846)*0.5;
    return (x+0.5)*log(x)-x+logpi;
}

/*---------------------------------------------------------------*
 |   PartiMove : A move set for Partition                        |
 |   USAGE : Anneal GoadFunction PartiMove IntVector Population  |
 *---------------------------------------------------------------*/
real Combination(Int n,Int k) {
    Int i,j;  real result,*Buf;
    Buf=(real*)GetMem(n,sizeof(real)); Buf[0]=1.0;
    for (i=0;i<=n;i++) for (j=i;j>=0;j--) Buf[j]+=Buf[j-1];
    result=Buf[k]; FreeMem(Buf); return result;
}

void PartiMove(const char*Mode) {
    static wptr CurArg, PopArg;
    static Int  UndoFrom,UndoTo;
    static Int *CurCnf,Pop;
    Int i,j;
    /*---------------------------*
     |  Initialize Move__ set      |
     *---------------------------*/
    if (Mode==InitMove__) {
        CurArg=ReadWord();
        if (CurArg->AsIvec.type!=IsIvec)
            Error(CurArg->AsIvec.name,EJ("Bad type","不正な型"));
        PopArg=ReadWord();
        if (PopArg->AsInt.type!=IsInt)
            Error(PopArg->AsInt.name,EJ("Bad type","不正な型"));
        Alloc(CurArg->AsIvec.name);
        CurCnf    =*CurArg->AsIvec.addr;
        Pop=*PopArg->AsInt.addr;
        CnfSize   =(Int)CurArg->AsIvec.row;
        GlbAcc99  =log(Combination(Pop+CnfSize-1,CnfSize-1));
        SpaceSize =2.0*CnfSize;
        Alloc("BestCnf");
        if (SpaceSize==0.0) Beta=1; else Beta=1/SpaceSize;
        LocAcc99  =-log(Beta);
    }
    /*-----------------------------*
     |  Generate__ a Random state  |
     *-----------------------------*/
    else if (Mode==Generate__) {
        for (i=0;i<CnfSize;i++) CurCnf[i]=0;
        for (i=0;i<Pop;i++) CurCnf[(size_t)(Rand()%CnfSize)]++;
    }
    /*--------------------------*
     |  Select a Random Move__  |
     *--------------------------*/
    else if (Mode==Move__) {
        do i=(Int)(Rand()%CnfSize); while (CurCnf[i]==0);
        UndoFrom=i;
        do j=(Int)(Rand()%CnfSize); while (i==j);
        UndoTo=j;
        CurCnf[i]--;  CurCnf[j]++;
    }
    /*---------------------------*
     |  Accept Best state        |
     |  Display Best state       |
     |  Load Best state          |
     |  Undo the previous move   |
     *---------------------------*/
    else if (Mode==AcceptBest__)
        for(i=0;i<CnfSize;i++) BestCnf[i]=CurCnf[i];
    else if (Mode==DisplayBest__)
        Print(CurArg->AsIvec.name);
    else if (Mode==LoadBest__)
        for(i=0;i<CnfSize;i++) CurCnf[i]=BestCnf[i];
    else if (Mode==UndoMove__)
    { CurCnf[UndoFrom]++; CurCnf[UndoTo]--; }
    else Error("PartiMove",
              EJ("Use Anneal only","アニーリングでのみ使用可"));
}

/*--------------------------------------------------------------*
 |   WalkMove : A move set for multi dimensional random walk    |
 |   USAGE :    WalkMove IntVector LowerBound UpperBound        |
 |              Anneal GoadFunction WalkMove                    |
 *--------------------------------------------------------------*/
void WalkMove(const char*Mode) {
    static Int UndoPos,UndoNum;
    static Int *CurCnf,*MinCnf,*MaxCnf;
    static wptr CurArg,MinArg,MaxArg;
    Int i,j;
    /*---------------------------*
     |  Initialize Move__ set      |
     *---------------------------*/
    if (Mode==InitMove__) {
        CurArg=ReadWord(); MinArg=ReadWord(); MaxArg=ReadWord();
        if (CurArg->AsIvec.type!=IsIvec)
            Error(CurArg->AsIvec.name,EJ("Bad type","不正な型"));
        if (MinArg->AsIvec.type!=IsIvec)
            Error(MinArg->AsIvec.name,EJ("Bad type","不正な型"));
        if (MaxArg->AsIvec.type!=IsIvec)
            Error(MaxArg->AsIvec.name,EJ("Bad type","不正な型"));
        Check(MinArg->AsIvec.name);
        Check(MaxArg->AsIvec.name);
        if (MinArg->AsIvec.row_!=CurArg->AsIvec.row_)
            Error(MinArg->AsIvec.name,EJ("Bad size","サイズが違う"));
        if (MaxArg->AsIvec.row_!=CurArg->AsIvec.row_)
            Error(MaxArg->AsIvec.name,EJ("Bad size","サイズが違う"));
        if (MinArg==CurArg)
            Error(MinArg->AsIvec.name,EJ("The same variable","同じ変数"));
        if (MaxArg==CurArg||MaxArg==MinArg)
            Error(MaxArg->AsIvec.name,EJ("The same variable","同じ変数"));
        Alloc(CurArg->AsIvec.name);
        CurCnf=*CurArg->AsIvec.addr;   MinCnf=*MinArg->AsIvec.addr;
        MaxCnf=*MaxArg->AsIvec.addr;   CnfSize=(Int)CurArg->AsIvec.row;
        GlbAcc99=0.0; SpaceSize=0.0;
        Alloc("BestCnf");
        for (i=0;i<CnfSize;i++)
            if (MaxCnf[i]>MinCnf[i]) {
                SpaceSize+=2;
                GlbAcc99 +=log(1.0+MaxCnf[i]-MinCnf[i]);
            }else if (MaxCnf[i]<MinCnf[i]) 
                Error(MaxArg->AsIvec.name,EJ("Max is smaller than Min",
                                     "最大値が最小値よりも小さい"));
        if (SpaceSize==0.0) Beta=1; else Beta=1/SpaceSize;
        LocAcc99=-log(Beta);
    }
    /*-----------------------------*
     |  Generate__ a Random state  |
     *-----------------------------*/
    else if (Mode==Generate__) {
        for (i=0;i<CnfSize;i++) 
         if (MaxCnf[i]>MinCnf[i])
             CurCnf[i]=MinCnf[i]+
                        (Int)(Rand()%(MaxCnf[i]-MinCnf[i]+1));
          else CurCnf[i]=MinCnf[i];
    }
    /*------------------------*
     |  Select a Random Move__  |
     *------------------------*/
    else if (Mode==Move__) {
        do i=(Int)(Rand()%CnfSize); while (MaxCnf[i]<=MinCnf[i]);
        UndoPos=i;
        UndoNum=j=CurCnf[i];
        if (Rand()&1) j++;  else j--;
        if      (j>MaxCnf[i]) CurCnf[i]=MinCnf[i];
        else if (j<MinCnf[i]) CurCnf[i]=MaxCnf[i];
        else                  CurCnf[i]=j;
    }
    /*---------------------------*
     |  Accept Best state        |
     |  Display Best state       |
     |  Load Best state          |
     |  Undo the previous move   |
     *---------------------------*/
    else if (Mode==AcceptBest__)
        for(i=0;i<CnfSize;i++) BestCnf[i]=CurCnf[i];
    else if (Mode==DisplayBest__)
        Print(CurArg->AsIvec.name);
    else if (Mode==LoadBest__)
        for(i=0;i<CnfSize;i++) CurCnf[i]=BestCnf[i];
    else if (Mode==UndoMove__)
        CurCnf[UndoPos]=UndoNum;
    else Error("WalkMove",EJ("Use Anneal only","アニーリングでのみ使用可"));
}

/*--------------------------------------------------------------*
 |   PermuMove: A move set for travelling salesman problems.    |
 |   USAGE    : PermuMove NodeVec REV                           |
 |              Anneal GoadFunction PermuMove                   |
 |                     or                                       |
 |              PermuMove NodeVec  SWP                          |
 |              PermuMove NodeVec1 REV  NodeVec2 REV            |
 |              PermuMove NodeVec1 REV  NodeVec2 SWP            |
 |              PermuMove NodeVec1 SWP  NodeVec2 REV            |
 |              PermuMove NodeVec1 SWP  NodeVec2 SWP            |
 *--------------------------------------------------------------*/
typedef struct {
    Int *Config,*Best,Left,Right,Size;
    bool RevOrNot;
    wptr Nvec;
} sortrec;

void Swap(Int*NextOrder,Int A,Int B) {
    Int CityA,CityB;
    CityA=NextOrder[A];
    CityB=NextOrder[B];
    NextOrder[A]=CityB;
    NextOrder[B]=CityA;
}

void Reverse(Int*NextOrder,Int A,Int B) {
    Int C,CityA,CityB;
    if (A>B) { C=A; A=B; B=C; }
    while (A<B) {
        CityA=NextOrder[A];
        CityB=NextOrder[B];
        NextOrder[A++]=CityB;
        NextOrder[B--]=CityA;
    }
}

void ReadPermu(sortrec*p) {
    char*Switch;  Int i;
    p->Nvec=ReadWord();
    if (p->Nvec->AsCvec.type!=IsCvec)
        Error(p->Nvec->AsCvec.name,EJ("Bad type","不正な型"));
    if (p->Nvec->AsCvec.row_!=p->Nvec->AsCvec.def.ptr->row_)
        Error(p->Nvec->AsCvec.name,
        EJ("Different size","サイズが異なる"));
    Switch=ReadStr();
    if      (strcmp(Switch,"REV")==0) p->RevOrNot=true;
    else if (strcmp(Switch,"SWP")==0) p->RevOrNot=false;
    else Error(Switch,EJ("SWP or REV expected",
        "REV か SWP を指定すること"));
    Alloc(p->Nvec->AsCvec.name);
    p->Config=*p->Nvec->AsCvec.addr;
    p->Size=(Int)p->Nvec->AsCvec.row;
    FreeMem(p->Best);
    p->Best=(Int*)GetMem(p->Size,sizeof(Int));
    p->Config=*p->Nvec->AsCvec.addr;
    for (i=0;i<p->Size;i++)  p->Config[i]=i;
    SpaceSize+=0.5*p->Size*p->Size;
    GlbAcc99 +=stirling(p->Size);
    if (SpaceSize==0.0) Beta=1; else Beta=1/SpaceSize;
    LocAcc99=-log(Beta);
}

void PermuMove(const char*Mode) {
    static sortrec Rec1,Rec2; Int i,j,k;
    /*-------------------------------------------*
     |  Initialize Move__ set                    |
     |  PermuMove Cur <REV|SWP> [Cur2 <REV|SWP>] |
     *-------------------------------------------*/
    if (Mode==InitMove__) {
        SpaceSize=0.0;  GlbAcc99=0.0;
        ReadPermu(&Rec1);
        if (EndOfLine) Rec2.Nvec=0; else ReadPermu(&Rec2);
        if (Rec1.Nvec==Rec2.Nvec)
            Error(Rec1.Nvec->AsCvec.name,
                 EJ("The same variable","同じ変数"));
    }
    /*-----------------------------*
     |  Generate__ a Random state  |
     *-----------------------------*/
    else if (Mode==Generate__) {
        for (k=0;k<Rec1.Size;k++) {
            i=(Int)(Rand()%Rec1.Size); j=(Int)(Rand()%Rec1.Size);
            Swap(Rec1.Config,i,j);
        }
        if (Rec2.Nvec)  for (k=0;k<Rec2.Size;k++) {
            i=(Int)(Rand()%Rec1.Size); j=(Int)(Rand()%Rec1.Size);
            Swap(Rec2.Config,i,j);
        }
    }
    /*--------------------------*
     |  Select a Random Move__  |
     *--------------------------*/
   else if (Mode==Move__) {
       i=(Int)(Rand()%Rec1.Size);
       do j=(Int)(Rand()%Rec1.Size); while (j==i);
       Rec1.Left=i;  Rec1.Right=j;
       if (Rec1.RevOrNot) Reverse(Rec1.Config,i,j);
        else              Swap(   Rec1.Config,i,j);
       if (Rec2.Nvec) {
           i=(Int)(Rand()%Rec2.Size);
           do j=(Int)(Rand()%Rec2.Size); while (j==i);
           Rec2.Left=i;  Rec2.Right=j;
           if (Rec2.RevOrNot) Reverse(Rec2.Config,i,j);
           else               Swap(   Rec2.Config,i,j);
       }
   }
    /*---------------------------*
     |  Undo the previous move   |
     *---------------------------*/
   else if (Mode==UndoMove__) {
       if (Rec1.RevOrNot) Reverse(Rec1.Config,Rec1.Left,Rec1.Right);
          else            Swap(Rec1.Config,Rec1.Left,Rec1.Right);
       if (Rec2.Nvec) {
           if(Rec2.RevOrNot)
                Reverse(Rec2.Config,Rec2.Left,Rec2.Right);
           else Swap(Rec2.Config,Rec2.Left,Rec2.Right);
       }
   }
    /*---------------------------*
     |  Accept Best state        |
     *---------------------------*/
    else if (Mode==AcceptBest__) {
        for (i=0;i<Rec1.Size;i++) Rec1.Best[i]=Rec1.Config[i];
        if (Rec2.Nvec) for (i=0;i<Rec2.Size;i++)
            Rec2.Best[i]=Rec2.Config[i];
    }
     /*---------------------------*
      |  Display Best state       |
      *---------------------------*/
    else if (Mode==DisplayBest__) {
      Print(Rec1.Nvec->AsCvec.name);
      if (Rec2.Nvec) Print(Rec2.Nvec->AsCvec.name);
    }
    /*---------------------------*
     |  Load Best state          |
     *---------------------------*/
    else if (Mode==LoadBest__) {
        for (i=0;i<Rec1.Size;i++) Rec1.Config[i]=Rec1.Best[i];
        FreeMem(Rec1.Best); Rec1.Best=0;
        if (Rec2.Nvec) {
            for (i=0;i<Rec2.Size;i++) Rec2.Config[i]=Rec2.Best[i];
            FreeMem(Rec2.Best); Rec2.Best=0;
        }
    }
    else Error("PermuMove",
             EJ("Use Anneal only","アニーリングでのみ使用可"));
}

    /*--------------------------------------*
     |                                      |
     |       THE ANNEALING ALGORITHM        |
     |                                      |
     *--------------------------------------*/
func MoveSet,Epsilon,
     AvailMoveList[]={ WalkMove,PermuMove,PartiMove,0 };

#if FLOAT_REAL
real expon(real x) {
    if (x<-84) return 0;
    if (x> 84) return exp(84.0);
    return exp(x);
}
#else
real expon(real x) {
    if (x<-200.0) return 0.0;
    if (x> 200.0) return exp(200.0);
    return exp(x);
}
#endif

void TestGoalFunc(void) {
    static bool ScoreIsTheSame;
    static real FirstScore;
    if (Count<=10.0) {
       if (Count==1.0) {
           ScoreIsTheSame=true;
           FirstScore=CurrScore;
        }
       if (FirstScore!=CurrScore) ScoreIsTheSame=false;
       if (Count==10.0&&ScoreIsTheSame)
           Error(EpsiArg__->AsProc.name,
               EJ("Goal not change CurrScore","目的関数が変化しない"));
    } 
}

void InitAnnealing(void) {
    Int i;  real Sigma2,Sum=0.0,SqSum=0.0,n;
    LoopSize=Sample; n=LoopSize;
    for (i=0;i<LoopSize;i++) {
        (*MoveSet)(Generate__);
        (*Epsilon)(CalcCurrScore__);
        Count+=1.0;
        if (*EpsiList) Print(EpsiList);
        TestGoalFunc();
        if (CurrScore<BestScore) {
            (*MoveSet)(AcceptBest__);
            BestScore=CurrScore;
            if (DispBest) (*MoveSet)(DisplayBest__);
        }
        Sum+=CurrScore;
        SqSum+=CurrScore*CurrScore;
        if (CurrScore<=OptiScore) return;
    }
    Energy=Energy99=Sum/n;
    Sigma2=(SqSum-Sum*Sum/n)/(n-1);
    if (Sigma2>0) Sigma99=sqrt(Sigma2);  else Sigma99=0;
    Sigma=Sigma99; Temp =Sigma99*sqrt(1/Gamma);
}

void Metropolis(void) {
    real Sum=0,SqSum=0,OldEnergy=Energy,a=0,up=0,xlnx,Sigma2;
    real PrevScore,K,AvailLoop=MaxCount-Count; Int i,PrevSize;
    if (AvailLoop<=0.0)          { LoopSize=0; return; };
    if (AvailLoop<LoopSize)  LoopSize=(Int)AvailLoop;
    K=LoopSize;
    for (i=1;i<=LoopSize;i++) {
        PrevScore = CurrScore;
        (*MoveSet)(Move__);
        (*Epsilon)(CalcCurrScore__);
        if (  PrevScore>CurrScore||Unif()
            < expon((PrevScore-CurrScore)/Temp))
            Accept=true;  else Accept=false;
        Count+=1.0;
        if (*EpsiList) Print(EpsiList);
        if (Accept) {
            a+=1/K;
            if (PrevScore<CurrScore)
                up+=(CurrScore-PrevScore)/K;
        }else{
            CurrScore = PrevScore;
            (*MoveSet)(UndoMove__);
        }
        if (CurrScore<BestScore) {
            BestScore=CurrScore;
            (*MoveSet)(AcceptBest__);
            if (DispBest) (*MoveSet)(DisplayBest__);
        }
        Sum+=CurrScore;
        SqSum+=CurrScore*CurrScore;
        if (BestScore<=OptiScore) return;
    }
    if (*MetList) Print(MetList);
    Energy=Sum/K;
    GlbAcc-=(OldEnergy-Energy)/Temp;
    Sigma2=(SqSum-Sum*Sum/K)/(K-0.999);
    if (Sigma2>0)  Sigma=sqrt(Sigma2);  else Sigma=0;
    if (0.9999<a&&a<1.0001) xlnx = 0;
    else xlnx = (1-a) * log(1-a);
    LocAcc = (LocAcc + a*log(1/Beta)+up/Temp-xlnx)/2;
    if (GlbAcc*Multi>LocAcc*MaxLoop)  LocAcc=GlbAcc*Multi/MaxLoop;
    PrevSize = LoopSize;
    LoopSize = (Int)(GlbAcc/LocAcc*Multi+0.5);
    if (LoopSize<PrevSize) LoopSize=PrevSize;
}

void ReadEpsiAndMove(void) {
    func*f;
    Count=0; BestScore=1.0e37;
    EpsiArg__=ReadWord(); MoveArg__=ReadWord();
    if (EpsiArg__->AsProc.type>IsProc)
        Error(EpsiArg__->AsProc.name,EJ("Bad type","不正な型"));
    if (MoveArg__->AsProc.type>IsProc)
        Error(MoveArg__->AsProc.name,EJ("Bad type","不正な型"));
    Epsilon=EpsiArg__->AsFunc.addr; MoveSet=MoveArg__->AsFunc.addr;
    for (f=AvailMoveList;*f;f++)  if (*f==MoveSet) break;
    if (*f==0) Error(MoveArg__->AsFunc.name,
                     EJ("Move expected","Moveが必要"));
    (*MoveSet)(InitMove__);
}

void Anneal(const char*argument) {
    real Te,TT,Mu,Dt=0.0; bool TempFlag,MaxTeFlag;
    CopyArg("Anneal",argument); ReadEpsiAndMove();
    EndArg(); InitAnnealing();
    if (BestScore<=OptiScore) goto QuitAnneal;
    Mu=Sigma99;   LocAcc=LocAcc99;   GlbAcc=GlbAcc99;  MaxTe=1e-37;
    if (*InitList) Print(InitList);
    LoopSize=(Int)(GlbAcc99/LocAcc99*Multi+0.5);
    if (LoopSize<MinLoop) LoopSize=MinLoop;
    do {
         Metropolis();
         if (BestScore<=OptiScore)  goto QuitAnneal;
         if (LoopSize==0) goto QuitAnneal;
         Mu    = Smooth*Sigma+(1-Smooth)*Mu*Temp/(Temp+Dt);
         Sigma = Mu;
         if (2*Gamma*Temp>Sigma) Dt=Temp/2;
         else                      Dt=Gamma*Temp*Temp/Sigma;
         TT    = Sigma99/Sigma*Temp;
         Temp -= Dt;
         Te    = TT*expon(-(TT*TT)*GlbAcc99/(Sigma99*Sigma99)+0.5);
         if (Te>MaxTe) MaxTe=Te;
         TempFlag=(bool)(Sigma*Sigma>Crit*Temp*(Energy99-Energy));
         MaxTeFlag=(bool)(Temp>MaxTe);
    } while (TempFlag||MaxTeFlag);
    Temp/=100.0;
    Metropolis();
 QuitAnneal:
    (*MoveSet)(LoadBest__);
}

void Improve(const char*argument) {
    CopyArg("Improve",argument);
    ReadEpsiAndMove(); EndArg();
    for (Count=0;Count<ImpSamp+ImpLoop;Count+=1.0) {
        if (Count==ImpSamp) (*MoveSet)(LoadBest__);
        if (Count<ImpSamp)  (*MoveSet)(Generate__);
          else              (*MoveSet)(Move__);
        (*Epsilon)(CalcCurrScore__);
        if (*EpsiList) Print(EpsiList);
        TestGoalFunc();
        if (CurrScore<BestScore) {
            (*MoveSet)(AcceptBest__);
            BestScore = CurrScore;
            if (DispBest) (*MoveSet)(DisplayBest__);
        }
        if (CurrScore<=OptiScore) break;
     }
    (*MoveSet)(LoadBest__);
}
#endif
